@@ -17,6 +17,7 @@
 package org.apache.catalina.filters;
 
 import java.io.IOException;
+import java.io.ObjectInputStream;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
@@ -26,9 +27,8 @@ import java.util.List;
 import java.util.Locale;
 import java.util.Set;
 
-import javax.servlet.Filter;
 import javax.servlet.FilterChain;
-import javax.servlet.FilterConfig;
+import javax.servlet.GenericFilter;
 import javax.servlet.ServletException;
 import javax.servlet.ServletRequest;
 import javax.servlet.ServletResponse;
@@ -43,9 +43,10 @@ import org.apache.tomcat.util.res.StringManager;
 
 /**
  * <p>
- * A {@link Filter} that enable client-side cross-origin requests by implementing W3C's CORS (<b>C</b>ross-<b>O</b>rigin
- * <b>R</b>esource <b>S</b>haring) specification for resources. Each {@link HttpServletRequest} request is inspected as
- * per specification, and appropriate response headers are added to {@link HttpServletResponse}.
+ * A {@link javax.servlet.Filter} that enable client-side cross-origin requests by implementing W3C's CORS
+ * (<b>C</b>ross-<b>O</b>rigin <b>R</b>esource <b>S</b>haring) specification for resources. Each
+ * {@link HttpServletRequest} request is inspected as per specification, and appropriate response headers are added to
+ * {@link HttpServletResponse}.
  * </p>
  * <p>
  * By default, it also sets following request attributes, that help to determine the nature of the request downstream.
@@ -72,11 +73,13 @@ import org.apache.tomcat.util.res.StringManager;
  *
  * @see <a href="http://www.w3.org/TR/cors/">CORS specification</a>
  */
-public class CorsFilter implements Filter {
+public class CorsFilter extends GenericFilter {
 
-    private final Log log = LogFactory.getLog(CorsFilter.class); // must not be static
+    private static final long serialVersionUID = 1L;
     private static final StringManager sm = StringManager.getManager(CorsFilter.class);
 
+    private transient Log log = LogFactory.getLog(CorsFilter.class); // must not be static
+
 
     /**
      * A {@link Collection} of origins consisting of zero or more origins that are allowed access to the resource.
@@ -137,7 +140,7 @@ public class CorsFilter implements Filter {
 
         // Adds CORS specific attributes to request.
         if (isDecorateRequest()) {
-            CorsFilter.decorateCORSProperties(request, requestType);
+            decorateCORSProperties(request, requestType);
         }
         switch (requestType) {
             case SIMPLE:
@@ -163,33 +166,28 @@ public class CorsFilter implements Filter {
 
 
     @Override
-    public void init(final FilterConfig filterConfig) throws ServletException {
-        parseAndStore(getInitParameter(filterConfig, PARAM_CORS_ALLOWED_ORIGINS, DEFAULT_ALLOWED_ORIGINS),
-                getInitParameter(filterConfig, PARAM_CORS_ALLOWED_METHODS, DEFAULT_ALLOWED_HTTP_METHODS),
-                getInitParameter(filterConfig, PARAM_CORS_ALLOWED_HEADERS, DEFAULT_ALLOWED_HTTP_HEADERS),
-                getInitParameter(filterConfig, PARAM_CORS_EXPOSED_HEADERS, DEFAULT_EXPOSED_HEADERS),
-                getInitParameter(filterConfig, PARAM_CORS_SUPPORT_CREDENTIALS, DEFAULT_SUPPORTS_CREDENTIALS),
-                getInitParameter(filterConfig, PARAM_CORS_PREFLIGHT_MAXAGE, DEFAULT_PREFLIGHT_MAXAGE),
-                getInitParameter(filterConfig, PARAM_CORS_REQUEST_DECORATE, DEFAULT_DECORATE_REQUEST));
+    public void init() throws ServletException {
+        parseAndStore(getInitParameter(PARAM_CORS_ALLOWED_ORIGINS, DEFAULT_ALLOWED_ORIGINS),
+                getInitParameter(PARAM_CORS_ALLOWED_METHODS, DEFAULT_ALLOWED_HTTP_METHODS),
+                getInitParameter(PARAM_CORS_ALLOWED_HEADERS, DEFAULT_ALLOWED_HTTP_HEADERS),
+                getInitParameter(PARAM_CORS_EXPOSED_HEADERS, DEFAULT_EXPOSED_HEADERS),
+                getInitParameter(PARAM_CORS_SUPPORT_CREDENTIALS, DEFAULT_SUPPORTS_CREDENTIALS),
+                getInitParameter(PARAM_CORS_PREFLIGHT_MAXAGE, DEFAULT_PREFLIGHT_MAXAGE),
+                getInitParameter(PARAM_CORS_REQUEST_DECORATE, DEFAULT_DECORATE_REQUEST));
     }
 
 
     /**
      * This method returns the parameter's value if it exists, or defaultValue if not.
      *
-     * @param filterConfig The configuration for the filter
      * @param name         The parameter's name
      * @param defaultValue The default value to return if the parameter does not exist
      *
      * @return The parameter's value or the default value if the parameter does not exist
      */
-    private String getInitParameter(FilterConfig filterConfig, String name, String defaultValue) {
+    private String getInitParameter(String name, String defaultValue) {
 
-        if (filterConfig == null) {
-            return defaultValue;
-        }
-
-        String value = filterConfig.getInitParameter(name);
+        String value = getInitParameter(name);
         if (value != null) {
             return value;
         }
@@ -220,7 +218,7 @@ public class CorsFilter implements Filter {
                     CorsFilter.CORSRequestType.ACTUAL));
         }
 
-        final String origin = request.getHeader(CorsFilter.REQUEST_HEADER_ORIGIN);
+        final String origin = request.getHeader(REQUEST_HEADER_ORIGIN);
         final String method = request.getMethod();
 
         // Section 6.1.2
@@ -260,7 +258,7 @@ public class CorsFilter implements Filter {
                     CORSRequestType.PRE_FLIGHT.name().toLowerCase(Locale.ENGLISH)));
         }
 
-        final String origin = request.getHeader(CorsFilter.REQUEST_HEADER_ORIGIN);
+        final String origin = request.getHeader(REQUEST_HEADER_ORIGIN);
 
         // Section 6.2.2
         if (!isOriginAllowed(origin)) {
@@ -269,7 +267,7 @@ public class CorsFilter implements Filter {
         }
 
         // Section 6.2.3
-        String accessControlRequestMethod = request.getHeader(CorsFilter.REQUEST_HEADER_ACCESS_CONTROL_REQUEST_METHOD);
+        String accessControlRequestMethod = request.getHeader(REQUEST_HEADER_ACCESS_CONTROL_REQUEST_METHOD);
         if (accessControlRequestMethod == null) {
             handleInvalidCORS(request, response, filterChain);
             return;
@@ -278,8 +276,7 @@ public class CorsFilter implements Filter {
         }
 
         // Section 6.2.4
-        String accessControlRequestHeadersHeader =
-                request.getHeader(CorsFilter.REQUEST_HEADER_ACCESS_CONTROL_REQUEST_HEADERS);
+        String accessControlRequestHeadersHeader = request.getHeader(REQUEST_HEADER_ACCESS_CONTROL_REQUEST_HEADERS);
         List<String> accessControlRequestHeaders = new ArrayList<>();
         if (accessControlRequestHeadersHeader != null && !accessControlRequestHeadersHeader.trim().isEmpty()) {
             String[] headers = accessControlRequestHeadersHeader.trim().split(",");
@@ -324,7 +321,11 @@ public class CorsFilter implements Filter {
     private void handleNonCORS(final HttpServletRequest request, final HttpServletResponse response,
             final FilterChain filterChain) throws IOException, ServletException {
 
-        addStandardHeaders(request, response);
+        if (!isAnyOriginAllowed()) {
+            // If only specific origins are allowed, the response will vary by
+            // origin
+            ResponseUtil.addVaryFieldName(response, REQUEST_HEADER_ORIGIN);
+        }
 
         // Let request pass.
         filterChain.doFilter(request, response);
@@ -340,7 +341,7 @@ public class CorsFilter implements Filter {
      */
     private void handleInvalidCORS(final HttpServletRequest request, final HttpServletResponse response,
             final FilterChain filterChain) {
-        String origin = request.getHeader(CorsFilter.REQUEST_HEADER_ORIGIN);
+        String origin = request.getHeader(REQUEST_HEADER_ORIGIN);
         String method = request.getMethod();
         String accessControlRequestHeaders = request.getHeader(REQUEST_HEADER_ACCESS_CONTROL_REQUEST_HEADERS);
 
@@ -369,7 +370,7 @@ public class CorsFilter implements Filter {
     private void addStandardHeaders(final HttpServletRequest request, final HttpServletResponse response) {
 
         final String method = request.getMethod();
-        final String origin = request.getHeader(CorsFilter.REQUEST_HEADER_ORIGIN);
+        final String origin = request.getHeader(REQUEST_HEADER_ORIGIN);
 
         // Local copy to avoid concurrency issues if isAnyOriginAllowed()
         // is overridden.
@@ -377,7 +378,7 @@ public class CorsFilter implements Filter {
         if (!anyOriginAllowed) {
             // If only specific origins are allowed, the response will vary by
             // origin
-            ResponseUtil.addVaryFieldName(response, CorsFilter.REQUEST_HEADER_ORIGIN);
+            ResponseUtil.addVaryFieldName(response, REQUEST_HEADER_ORIGIN);
         }
 
         // CORS requests (SIMPLE, ACTUAL, PRE_FLIGHT) set the following headers
@@ -388,18 +389,18 @@ public class CorsFilter implements Filter {
         // Add a single Access-Control-Allow-Origin header.
         if (anyOriginAllowed) {
             // If any origin is allowed, return header with '*'.
-            response.addHeader(CorsFilter.RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN, "*");
+            response.addHeader(RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN, "*");
         } else {
             // Add a single Access-Control-Allow-Origin header, with the value
             // of the Origin header as value.
-            response.addHeader(CorsFilter.RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN, origin);
+            response.addHeader(RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_ORIGIN, origin);
         }
 
         // If the resource supports credentials, add a single
         // Access-Control-Allow-Credentials header with the case-sensitive
         // string "true" as value.
         if (isSupportsCredentials()) {
-            response.addHeader(CorsFilter.RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_CREDENTIALS, "true");
+            response.addHeader(RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_CREDENTIALS, "true");
         }
 
         // If the list of exposed headers is not empty add one or more
@@ -408,17 +409,17 @@ public class CorsFilter implements Filter {
         // Local copy to avoid concurrency issues if getExposedHeaders()
         // is overridden.
         Collection<String> exposedHeaders = getExposedHeaders();
-        if ((exposedHeaders != null) && (exposedHeaders.size() > 0)) {
+        if (exposedHeaders != null && exposedHeaders.size() > 0) {
             String exposedHeadersString = join(exposedHeaders, ",");
-            response.addHeader(CorsFilter.RESPONSE_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS, exposedHeadersString);
+            response.addHeader(RESPONSE_HEADER_ACCESS_CONTROL_EXPOSE_HEADERS, exposedHeadersString);
         }
 
         if ("OPTIONS".equals(method)) {
             // For an OPTIONS request, the response will vary based on the
             // value or absence of the following headers. Hence they need be be
             // included in the Vary header.
-            ResponseUtil.addVaryFieldName(response, CorsFilter.REQUEST_HEADER_ACCESS_CONTROL_REQUEST_METHOD);
-            ResponseUtil.addVaryFieldName(response, CorsFilter.REQUEST_HEADER_ACCESS_CONTROL_REQUEST_HEADERS);
+            ResponseUtil.addVaryFieldName(response, REQUEST_HEADER_ACCESS_CONTROL_REQUEST_METHOD);
+            ResponseUtil.addVaryFieldName(response, REQUEST_HEADER_ACCESS_CONTROL_REQUEST_HEADERS);
 
             // CORS PRE_FLIGHT (OPTIONS) requests set the following headers although
             // non-CORS OPTIONS requests do not need to. The headers are always set
@@ -428,34 +429,26 @@ public class CorsFilter implements Filter {
             // is overridden.
             long preflightMaxAge = getPreflightMaxAge();
             if (preflightMaxAge > 0) {
-                response.addHeader(CorsFilter.RESPONSE_HEADER_ACCESS_CONTROL_MAX_AGE, String.valueOf(preflightMaxAge));
+                response.addHeader(RESPONSE_HEADER_ACCESS_CONTROL_MAX_AGE, String.valueOf(preflightMaxAge));
             }
 
             // Local copy to avoid concurrency issues if getAllowedHttpMethods()
             // is overridden.
             Collection<String> allowedHttpMethods = getAllowedHttpMethods();
-            if ((allowedHttpMethods != null) && (!allowedHttpMethods.isEmpty())) {
-                response.addHeader(CorsFilter.RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_METHODS,
-                        join(allowedHttpMethods, ","));
+            if (allowedHttpMethods != null && !allowedHttpMethods.isEmpty()) {
+                response.addHeader(RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_METHODS, join(allowedHttpMethods, ","));
             }
 
             // Local copy to avoid concurrency issues if getAllowedHttpHeaders()
             // is overridden.
             Collection<String> allowedHttpHeaders = getAllowedHttpHeaders();
-            if ((allowedHttpHeaders != null) && (!allowedHttpHeaders.isEmpty())) {
-                response.addHeader(CorsFilter.RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_HEADERS,
-                        join(allowedHttpHeaders, ","));
+            if (allowedHttpHeaders != null && !allowedHttpHeaders.isEmpty()) {
+                response.addHeader(RESPONSE_HEADER_ACCESS_CONTROL_ALLOW_HEADERS, join(allowedHttpHeaders, ","));
             }
         }
     }
 
 
-    @Override
-    public void destroy() {
-        // NOOP
-    }
-
-
     /**
      * Decorates the {@link HttpServletRequest}, with CORS attributes.
      * <ul>
@@ -484,26 +477,24 @@ public class CorsFilter implements Filter {
         switch (corsRequestType) {
             case SIMPLE:
             case ACTUAL:
-                request.setAttribute(CorsFilter.HTTP_REQUEST_ATTRIBUTE_IS_CORS_REQUEST, Boolean.TRUE);
-                request.setAttribute(CorsFilter.HTTP_REQUEST_ATTRIBUTE_ORIGIN,
-                        request.getHeader(CorsFilter.REQUEST_HEADER_ORIGIN));
-                request.setAttribute(CorsFilter.HTTP_REQUEST_ATTRIBUTE_REQUEST_TYPE,
+                request.setAttribute(HTTP_REQUEST_ATTRIBUTE_IS_CORS_REQUEST, Boolean.TRUE);
+                request.setAttribute(HTTP_REQUEST_ATTRIBUTE_ORIGIN, request.getHeader(REQUEST_HEADER_ORIGIN));
+                request.setAttribute(HTTP_REQUEST_ATTRIBUTE_REQUEST_TYPE,
                         corsRequestType.name().toLowerCase(Locale.ENGLISH));
                 break;
             case PRE_FLIGHT:
-                request.setAttribute(CorsFilter.HTTP_REQUEST_ATTRIBUTE_IS_CORS_REQUEST, Boolean.TRUE);
-                request.setAttribute(CorsFilter.HTTP_REQUEST_ATTRIBUTE_ORIGIN,
-                        request.getHeader(CorsFilter.REQUEST_HEADER_ORIGIN));
-                request.setAttribute(CorsFilter.HTTP_REQUEST_ATTRIBUTE_REQUEST_TYPE,
+                request.setAttribute(HTTP_REQUEST_ATTRIBUTE_IS_CORS_REQUEST, Boolean.TRUE);
+                request.setAttribute(HTTP_REQUEST_ATTRIBUTE_ORIGIN, request.getHeader(REQUEST_HEADER_ORIGIN));
+                request.setAttribute(HTTP_REQUEST_ATTRIBUTE_REQUEST_TYPE,
                         corsRequestType.name().toLowerCase(Locale.ENGLISH));
                 String headers = request.getHeader(REQUEST_HEADER_ACCESS_CONTROL_REQUEST_HEADERS);
                 if (headers == null) {
                     headers = "";
                 }
-                request.setAttribute(CorsFilter.HTTP_REQUEST_ATTRIBUTE_REQUEST_HEADERS, headers);
+                request.setAttribute(HTTP_REQUEST_ATTRIBUTE_REQUEST_HEADERS, headers);
                 break;
             case NOT_CORS:
-                request.setAttribute(CorsFilter.HTTP_REQUEST_ATTRIBUTE_IS_CORS_REQUEST, Boolean.FALSE);
+                request.setAttribute(HTTP_REQUEST_ATTRIBUTE_IS_CORS_REQUEST, Boolean.FALSE);
                 break;
             default:
                 // Don't set any attributes
@@ -832,6 +823,16 @@ public class CorsFilter implements Filter {
     }
 
 
+    /*
+     * Log objects are not Serializable but this Filter is because it extends GenericFilter. Tomcat won't serialize a
+     * Filter but in case something else does...
+     */
+    private void readObject(ObjectInputStream ois) throws ClassNotFoundException, IOException {
+        ois.defaultReadObject();
+        log = LogFactory.getLog(CorsFilter.class);
+    }
+
+
     // -------------------------------------------------- CORS Response Headers
 
     /**
