@@ -16,25 +16,32 @@
  */
 package org.apache.coyote.http2;
 
+import java.io.BufferedReader;
+import java.io.ByteArrayInputStream;
 import java.io.IOException;
+import java.io.InputStreamReader;
+import java.io.Reader;
 import java.nio.ByteBuffer;
 import java.nio.charset.StandardCharsets;
 
+import javax.servlet.http.WebConnection;
+
 import org.apache.coyote.ProtocolException;
 import org.apache.coyote.http2.HpackDecoder.HeaderEmitter;
 import org.apache.juli.logging.Log;
 import org.apache.juli.logging.LogFactory;
 import org.apache.tomcat.util.buf.ByteBufferUtils;
+import org.apache.tomcat.util.http.parser.Priority;
 import org.apache.tomcat.util.res.StringManager;
 
 class Http2Parser {
 
-    private static final Log log = LogFactory.getLog(Http2Parser.class);
-    private static final StringManager sm = StringManager.getManager(Http2Parser.class);
+    protected static final Log log = LogFactory.getLog(Http2Parser.class);
+    protected static final StringManager sm = StringManager.getManager(Http2Parser.class);
 
     static final byte[] CLIENT_PREFACE_START = "PRI * HTTP/2.0\r\n\r\nSM\r\n\r\n".getBytes(StandardCharsets.ISO_8859_1);
 
-    private final String connectionId;
+    protected final String connectionId;
     protected final Input input;
     private final Output output;
     private final byte[] frameHeaderBuffer = new byte[9];
@@ -81,7 +88,7 @@ class Http2Parser {
     }
 
 
-    private boolean readFrame(boolean block, FrameType expected) throws IOException, Http2Exception {
+    protected boolean readFrame(boolean block, FrameType expected) throws IOException, Http2Exception {
 
         if (!input.fill(block, frameHeaderBuffer)) {
             return false;
@@ -96,50 +103,53 @@ class Http2Parser {
         try {
             validateFrame(expected, frameType, streamId, flags, payloadSize);
         } catch (StreamException se) {
-            swallowPayload(streamId, frameTypeId, payloadSize, false);
+            swallowPayload(streamId, frameTypeId, payloadSize, false, null);
             throw se;
         }
 
         switch (frameType) {
             case DATA:
-                readDataFrame(streamId, flags, payloadSize);
+                readDataFrame(streamId, flags, payloadSize, null);
                 break;
             case HEADERS:
-                readHeadersFrame(streamId, flags, payloadSize);
+                readHeadersFrame(streamId, flags, payloadSize, null);
                 break;
             case PRIORITY:
-                readPriorityFrame(streamId);
+                readPriorityFrame(streamId, null);
                 break;
             case RST:
-                readRstFrame(streamId);
+                readRstFrame(streamId, null);
                 break;
             case SETTINGS:
-                readSettingsFrame(flags, payloadSize);
+                readSettingsFrame(flags, payloadSize, null);
                 break;
             case PUSH_PROMISE:
-                readPushPromiseFrame(streamId, flags, payloadSize);
+                readPushPromiseFrame(streamId, flags, payloadSize, null);
                 break;
             case PING:
-                readPingFrame(flags);
+                readPingFrame(flags, null);
                 break;
             case GOAWAY:
-                readGoawayFrame(payloadSize);
+                readGoawayFrame(payloadSize, null);
                 break;
             case WINDOW_UPDATE:
-                readWindowUpdateFrame(streamId);
+                readWindowUpdateFrame(streamId, null);
                 break;
             case CONTINUATION:
-                readContinuationFrame(streamId, flags, payloadSize);
+                readContinuationFrame(streamId, flags, payloadSize, null);
+                break;
+            case PRIORITY_UPDATE:
+                readPriorityUpdateFrame(payloadSize, null);
                 break;
             case UNKNOWN:
-                readUnknownFrame(streamId, frameTypeId, flags, payloadSize);
+                readUnknownFrame(streamId, frameTypeId, flags, payloadSize, null);
         }
 
         return true;
     }
 
-
-    private void readDataFrame(int streamId, int flags, int payloadSize) throws Http2Exception, IOException {
+    protected void readDataFrame(int streamId, int flags, int payloadSize, ByteBuffer buffer)
+            throws Http2Exception, IOException {
         // Process the Stream
         int padLength = 0;
 
@@ -147,9 +157,13 @@ class Http2Parser {
 
         int dataLength;
         if (Flags.hasPadding(flags)) {
-            byte[] b = new byte[1];
-            input.fill(true, b);
-            padLength = b[0] & 0xFF;
+            if (buffer == null) {
+                byte[] b = new byte[1];
+                input.fill(true, b);
+                padLength = b[0] & 0xFF;
+            } else {
+                padLength = buffer.get() & 0xFF;
+            }
 
             if (padLength >= payloadSize) {
                 throw new ConnectionException(
@@ -176,11 +190,11 @@ class Http2Parser {
 
         ByteBuffer dest = output.startRequestBodyFrame(streamId, payloadSize, endOfStream);
         if (dest == null) {
-            swallowPayload(streamId, FrameType.DATA.getId(), dataLength, false);
+            swallowPayload(streamId, FrameType.DATA.getId(), dataLength, false, buffer);
             // Process padding before sending any notifications in case padding
             // is invalid.
             if (Flags.hasPadding(flags)) {
-                swallowPayload(streamId, FrameType.DATA.getId(), padLength, true);
+                swallowPayload(streamId, FrameType.DATA.getId(), padLength, true, buffer);
             }
             if (endOfStream) {
                 output.receivedEndOfStream(streamId);
@@ -189,18 +203,25 @@ class Http2Parser {
             synchronized (dest) {
                 if (dest.remaining() < payloadSize) {
                     // Client has sent more data than permitted by Window size
-                    swallowPayload(streamId, FrameType.DATA.getId(), dataLength, false);
+                    swallowPayload(streamId, FrameType.DATA.getId(), dataLength, false, buffer);
                     if (Flags.hasPadding(flags)) {
-                        swallowPayload(streamId, FrameType.DATA.getId(), padLength, true);
+                        swallowPayload(streamId, FrameType.DATA.getId(), padLength, true, buffer);
                     }
                     throw new StreamException(sm.getString("http2Parser.processFrameData.window", connectionId),
                             Http2Error.FLOW_CONTROL_ERROR, streamId);
                 }
-                input.fill(true, dest, dataLength);
+                if (buffer == null) {
+                    input.fill(true, dest, dataLength);
+                } else {
+                    int oldLimit = buffer.limit();
+                    buffer.limit(buffer.position() + dataLength);
+                    dest.put(buffer);
+                    buffer.limit(oldLimit);
+                }
                 // Process padding before sending any notifications in case
                 // padding is invalid.
                 if (Flags.hasPadding(flags)) {
-                    swallowPayload(streamId, FrameType.DATA.getId(), padLength, true);
+                    swallowPayload(streamId, FrameType.DATA.getId(), padLength, true, buffer);
                 }
                 if (endOfStream) {
                     output.receivedEndOfStream(streamId);
@@ -211,7 +232,8 @@ class Http2Parser {
     }
 
 
-    private void readHeadersFrame(int streamId, int flags, int payloadSize) throws Http2Exception, IOException {
+    protected void readHeadersFrame(int streamId, int flags, int payloadSize, ByteBuffer buffer)
+            throws Http2Exception, IOException {
 
         headersEndStream = Flags.isEndOfStream(flags);
 
@@ -221,7 +243,7 @@ class Http2Parser {
         try {
             hpackDecoder.setHeaderEmitter(output.headersStart(streamId, headersEndStream));
         } catch (StreamException se) {
-            swallowPayload(streamId, FrameType.HEADERS.getId(), payloadSize, false);
+            swallowPayload(streamId, FrameType.HEADERS.getId(), payloadSize, false, buffer);
             throw se;
         }
 
@@ -237,30 +259,29 @@ class Http2Parser {
         }
         if (optionalLen > 0) {
             byte[] optional = new byte[optionalLen];
-            input.fill(true, optional);
-            int optionalPos = 0;
+            if (buffer == null) {
+                input.fill(true, optional);
+            } else {
+                buffer.get(optional);
+            }
             if (padding) {
-                padLength = ByteUtil.getOneByte(optional, optionalPos++);
+                padLength = ByteUtil.getOneByte(optional, 0);
                 if (padLength >= payloadSize) {
                     throw new ConnectionException(sm.getString("http2Parser.processFrame.tooMuchPadding", connectionId,
                             Integer.toString(streamId), Integer.toString(padLength), Integer.toString(payloadSize)),
                             Http2Error.PROTOCOL_ERROR);
                 }
             }
-            if (priority) {
-                boolean exclusive = ByteUtil.isBit7Set(optional[optionalPos]);
-                int parentStreamId = ByteUtil.get31Bits(optional, optionalPos);
-                int weight = ByteUtil.getOneByte(optional, optionalPos + 4) + 1;
-                output.reprioritise(streamId, parentStreamId, exclusive, weight);
-            }
+
+            // Ignore RFC 7450 priority data if present
 
             payloadSize -= optionalLen;
             payloadSize -= padLength;
         }
 
-        readHeaderPayload(streamId, payloadSize);
+        readHeaderPayload(streamId, payloadSize, buffer);
 
-        swallowPayload(streamId, FrameType.HEADERS.getId(), padLength, true);
+        swallowPayload(streamId, FrameType.HEADERS.getId(), padLength, true, buffer);
 
         if (Flags.isEndOfHeaders(flags)) {
             onHeadersComplete(streamId);
@@ -270,26 +291,25 @@ class Http2Parser {
     }
 
 
-    private void readPriorityFrame(int streamId) throws Http2Exception, IOException {
-        byte[] payload = new byte[5];
-        input.fill(true, payload);
-
-        boolean exclusive = ByteUtil.isBit7Set(payload[0]);
-        int parentStreamId = ByteUtil.get31Bits(payload, 0);
-        int weight = ByteUtil.getOneByte(payload, 4) + 1;
-
-        if (streamId == parentStreamId) {
-            throw new StreamException(sm.getString("http2Parser.processFramePriority.invalidParent", connectionId,
-                    Integer.valueOf(streamId)), Http2Error.PROTOCOL_ERROR, streamId);
+    protected void readPriorityFrame(int streamId, ByteBuffer buffer) throws IOException {
+        // RFC 7450 priority frames are ignored. Still need to treat as overhead.
+        try {
+            swallowPayload(streamId, FrameType.PRIORITY.getId(), 5, false, buffer);
+        } catch (ConnectionException e) {
+            // Will never happen because swallowPayload() is called with isPadding set
+            // to false
         }
-
-        output.reprioritise(streamId, parentStreamId, exclusive, weight);
+        output.increaseOverheadCount(FrameType.PRIORITY);
     }
 
 
-    private void readRstFrame(int streamId) throws Http2Exception, IOException {
+    protected void readRstFrame(int streamId, ByteBuffer buffer) throws Http2Exception, IOException {
         byte[] payload = new byte[4];
-        input.fill(true, payload);
+        if (buffer == null) {
+            input.fill(true, payload);
+        } else {
+            buffer.get(payload);
+        }
 
         long errorCode = ByteUtil.getFourBytes(payload, 0);
         output.reset(streamId, errorCode);
@@ -298,7 +318,7 @@ class Http2Parser {
     }
 
 
-    private void readSettingsFrame(int flags, int payloadSize) throws Http2Exception, IOException {
+    protected void readSettingsFrame(int flags, int payloadSize, ByteBuffer buffer) throws Http2Exception, IOException {
         boolean ack = Flags.isAck(flags);
         if (payloadSize > 0 && ack) {
             throw new ConnectionException(sm.getString("http2Parser.processFrameSettings.ackWithNonZeroPayload"),
@@ -312,7 +332,11 @@ class Http2Parser {
             // Process the settings
             byte[] setting = new byte[6];
             for (int i = 0; i < payloadSize / 6; i++) {
-                input.fill(true, setting);
+                if (buffer == null) {
+                    input.fill(true, setting);
+                } else {
+                    buffer.get(setting);
+                }
                 int id = ByteUtil.getTwoBytes(setting, 0);
                 long value = ByteUtil.getFourBytes(setting, 2);
                 Setting key = Setting.valueOf(id);
@@ -334,28 +358,38 @@ class Http2Parser {
      * @param streamId    The pushed stream
      * @param flags       The flags set in the frame header
      * @param payloadSize The size of the payload in bytes
+     * @param buffer      The payload, if available
      *
      * @throws Http2Exception Always
      * @throws IOException    May be thrown by sub-classes that parse this frame
      */
-    protected void readPushPromiseFrame(int streamId, int flags, int payloadSize) throws Http2Exception, IOException {
+    protected void readPushPromiseFrame(int streamId, int flags, int payloadSize, ByteBuffer buffer)
+            throws Http2Exception, IOException {
         throw new ConnectionException(
                 sm.getString("http2Parser.processFramePushPromise", connectionId, Integer.valueOf(streamId)),
                 Http2Error.PROTOCOL_ERROR);
     }
 
 
-    private void readPingFrame(int flags) throws IOException {
+    protected void readPingFrame(int flags, ByteBuffer buffer) throws IOException {
         // Read the payload
         byte[] payload = new byte[8];
-        input.fill(true, payload);
+        if (buffer == null) {
+            input.fill(true, payload);
+        } else {
+            buffer.get(payload);
+        }
         output.pingReceive(payload, Flags.isAck(flags));
     }
 
 
-    private void readGoawayFrame(int payloadSize) throws IOException {
+    protected void readGoawayFrame(int payloadSize, ByteBuffer buffer) throws IOException {
         byte[] payload = new byte[payloadSize];
-        input.fill(true, payload);
+        if (buffer == null) {
+            input.fill(true, payload);
+        } else {
+            buffer.get(payload);
+        }
 
         int lastStreamId = ByteUtil.get31Bits(payload, 0);
         long errorCode = ByteUtil.getFourBytes(payload, 4);
@@ -367,9 +401,13 @@ class Http2Parser {
     }
 
 
-    private void readWindowUpdateFrame(int streamId) throws Http2Exception, IOException {
+    protected void readWindowUpdateFrame(int streamId, ByteBuffer buffer) throws Http2Exception, IOException {
         byte[] payload = new byte[4];
-        input.fill(true, payload);
+        if (buffer == null) {
+            input.fill(true, payload);
+        } else {
+            buffer.get(payload);
+        }
         int windowSizeIncrement = ByteUtil.get31Bits(payload, 0);
 
         if (log.isDebugEnabled()) {
@@ -392,7 +430,8 @@ class Http2Parser {
     }
 
 
-    private void readContinuationFrame(int streamId, int flags, int payloadSize) throws Http2Exception, IOException {
+    protected void readContinuationFrame(int streamId, int flags, int payloadSize, ByteBuffer buffer)
+            throws Http2Exception, IOException {
         if (headersCurrentStream == -1) {
             // No headers to continue
             throw new ConnectionException(sm.getString("http2Parser.processFrameContinuation.notExpected", connectionId,
@@ -405,7 +444,7 @@ class Http2Parser {
         // continuation frames
         output.headersContinue(payloadSize, endOfHeaders);
 
-        readHeaderPayload(streamId, payloadSize);
+        readHeaderPayload(streamId, payloadSize, buffer);
 
         if (endOfHeaders) {
             headersCurrentStream = -1;
@@ -414,7 +453,38 @@ class Http2Parser {
     }
 
 
-    protected void readHeaderPayload(int streamId, int payloadSize) throws Http2Exception, IOException {
+    protected void readPriorityUpdateFrame(int payloadSize, ByteBuffer buffer) throws Http2Exception, IOException {
+        // Identify prioritized stream ID
+        byte[] payload = new byte[payloadSize];
+        if (buffer == null) {
+            input.fill(true, payload);
+        } else {
+            buffer.get(payload);
+        }
+
+        int prioritizedStreamID = ByteUtil.get31Bits(payload, 0);
+
+        if (prioritizedStreamID == 0) {
+            throw new ConnectionException(sm.getString("http2Parser.processFramePriorityUpdate.streamZero"),
+                    Http2Error.PROTOCOL_ERROR);
+        }
+
+        ByteArrayInputStream bais = new ByteArrayInputStream(payload, 4, payloadSize - 4);
+        Reader r = new BufferedReader(new InputStreamReader(bais, StandardCharsets.US_ASCII));
+        Priority p = Priority.parsePriority(r);
+
+        if (log.isDebugEnabled()) {
+            log.debug(sm.getString("http2Parser.processFramePriorityUpdate.debug", connectionId,
+                    Integer.toString(prioritizedStreamID), Integer.toString(p.getUrgency()),
+                    Boolean.valueOf(p.getIncremental())));
+        }
+
+        output.priorityUpdate(prioritizedStreamID, p);
+    }
+
+
+    protected void readHeaderPayload(int streamId, int payloadSize, ByteBuffer buffer)
+            throws Http2Exception, IOException {
 
         if (log.isDebugEnabled()) {
             log.debug(sm.getString("http2Parser.processFrameHeaders.payload", connectionId, Integer.valueOf(streamId),
@@ -440,7 +510,14 @@ class Http2Parser {
             }
             int toRead = Math.min(headerReadBuffer.remaining(), remaining);
             // headerReadBuffer in write mode
-            input.fill(true, headerReadBuffer, toRead);
+            if (buffer == null) {
+                input.fill(true, headerReadBuffer, toRead);
+            } else {
+                int oldLimit = buffer.limit();
+                buffer.limit(buffer.position() + toRead);
+                headerReadBuffer.put(buffer);
+                buffer.limit(oldLimit);
+            }
             // switch to read mode
             headerReadBuffer.flip();
             try {
@@ -477,9 +554,10 @@ class Http2Parser {
     }
 
 
-    protected void readUnknownFrame(int streamId, int frameTypeId, int flags, int payloadSize) throws IOException {
+    protected void readUnknownFrame(int streamId, int frameTypeId, int flags, int payloadSize, ByteBuffer buffer)
+            throws IOException {
         try {
-            swallowPayload(streamId, frameTypeId, payloadSize, false);
+            swallowPayload(streamId, frameTypeId, payloadSize, false, buffer);
         } catch (ConnectionException e) {
             // Will never happen because swallowPayload() is called with isPadding set
             // to false
@@ -496,38 +574,46 @@ class Http2Parser {
      * @param frameTypeId Type of HTTP/2 frame for which the bytes will be swallowed
      * @param len         Number of bytes to swallow
      * @param isPadding   Are the bytes to be swallowed padding bytes?
+     * @param byteBuffer  Used with {@link Http2AsyncParser} to access the data that has already been read
      *
      * @throws IOException         If an I/O error occurs reading additional bytes into the input buffer.
      * @throws ConnectionException If the swallowed bytes are expected to have a value of zero but do not
      */
-    protected void swallowPayload(int streamId, int frameTypeId, int len, boolean isPadding)
+    protected void swallowPayload(int streamId, int frameTypeId, int len, boolean isPadding, ByteBuffer byteBuffer)
             throws IOException, ConnectionException {
         if (log.isDebugEnabled()) {
             log.debug(sm.getString("http2Parser.swallow.debug", connectionId, Integer.toString(streamId),
                     Integer.toString(len)));
         }
-        int read = 0;
-        int thisTime = 0;
         try {
             if (len == 0) {
                 return;
             }
-            byte[] buffer = new byte[1024];
-            while (read < len) {
-                thisTime = Math.min(buffer.length, len - read);
-                input.fill(true, buffer, 0, thisTime);
-                if (isPadding) {
-                    // Validate the padding is zero since receiving non-zero padding
-                    // is a strong indication of either a faulty client or a server
-                    // side bug.
-                    for (int i = 0; i < thisTime; i++) {
-                        if (buffer[i] != 0) {
-                            throw new ConnectionException(sm.getString("http2Parser.nonZeroPadding", connectionId,
-                                    Integer.toString(streamId)), Http2Error.PROTOCOL_ERROR);
+            if (!isPadding && byteBuffer != null) {
+                byteBuffer.position(byteBuffer.position() + len);
+            } else {
+                int read = 0;
+                byte[] buffer = new byte[1024];
+                while (read < len) {
+                    int thisTime = Math.min(buffer.length, len - read);
+                    if (byteBuffer == null) {
+                        input.fill(true, buffer, 0, thisTime);
+                    } else {
+                        byteBuffer.get(buffer, 0, thisTime);
+                    }
+                    if (isPadding) {
+                        // Validate the padding is zero since receiving non-zero padding
+                        // is a strong indication of either a faulty client or a server
+                        // side bug.
+                        for (int i = 0; i < thisTime; i++) {
+                            if (buffer[i] != 0) {
+                                throw new ConnectionException(sm.getString("http2Parser.nonZeroPadding", connectionId,
+                                        Integer.toString(streamId)), Http2Error.PROTOCOL_ERROR);
+                            }
                         }
                     }
+                    read += thisTime;
                 }
-                read += thisTime;
             }
         } finally {
             if (FrameType.DATA.getIdByte() == frameTypeId) {
@@ -540,12 +626,11 @@ class Http2Parser {
                     output.onSwallowedDataFramePayload(streamId, len);
                 }
             }
-            read += thisTime;
         }
     }
 
 
-    private void onHeadersComplete(int streamId) throws Http2Exception {
+    protected void onHeadersComplete(int streamId) throws Http2Exception {
         // Any left over data is a compression error
         if (headerReadBuffer.position() > 0) {
             throw new ConnectionException(sm.getString("http2Parser.processFrameHeaders.decodingDataLeft"),
@@ -557,11 +642,12 @@ class Http2Parser {
         // going to be thrown.
         hpackDecoder.getHeaderEmitter().validateHeaders();
 
-        output.headersEnd(streamId);
+        synchronized (output) {
+            output.headersEnd(streamId, headersEndStream);
 
-        if (headersEndStream) {
-            output.receivedEndOfStream(streamId);
-            headersEndStream = false;
+            if (headersEndStream) {
+                headersEndStream = false;
+            }
         }
 
         // Reset size for new request if the buffer was previously expanded
@@ -576,7 +662,7 @@ class Http2Parser {
      * validation should be performed in the appropriate readXxxFrame() method. For validation applicable to some but
      * not all frame types, use your judgement.
      */
-    private void validateFrame(FrameType expected, FrameType frameType, int streamId, int flags, int payloadSize)
+    protected void validateFrame(FrameType expected, FrameType frameType, int streamId, int flags, int payloadSize)
             throws Http2Exception {
 
         if (log.isDebugEnabled()) {
@@ -616,8 +702,11 @@ class Http2Parser {
 
     /**
      * Read and validate the connection preface from input using blocking IO.
+     *
+     * @param webConnection The connection
+     * @param stream        The current stream
      */
-    void readConnectionPreface() throws Http2Exception {
+    void readConnectionPreface(WebConnection webConnection, Stream stream) throws Http2Exception {
         byte[] data = new byte[CLIENT_PREFACE_START.length];
         try {
             input.fill(true, data);
@@ -656,9 +745,17 @@ class Http2Parser {
          */
         boolean fill(boolean block, byte[] data, int offset, int length) throws IOException;
 
-        boolean fill(boolean block, byte[] data) throws IOException;
+        default boolean fill(boolean block, byte[] data) throws IOException {
+            return fill(block, data, 0, data.length);
+        }
 
-        boolean fill(boolean block, ByteBuffer data, int len) throws IOException;
+        default boolean fill(boolean block, ByteBuffer data, int len) throws IOException {
+            boolean result = fill(block, data.array(), data.arrayOffset() + data.position(), len);
+            if (result) {
+                data.position(data.position() + len);
+            }
+            return result;
+        }
 
         int getMaxFrameSize();
     }
@@ -696,10 +793,7 @@ class Http2Parser {
 
         void headersContinue(int payloadSize, boolean endOfHeaders);
 
-        void headersEnd(int streamId) throws Http2Exception;
-
-        // Priority frames (also headers)
-        void reprioritise(int streamId, int parentStreamId, boolean exclusive, int weight) throws Http2Exception;
+        void headersEnd(int streamId, boolean endOfStream) throws Http2Exception;
 
         // Reset frames
         void reset(int streamId, long errorCode) throws Http2Exception;
@@ -718,6 +812,9 @@ class Http2Parser {
         // Window size
         void incrementWindowSize(int streamId, int increment) throws Http2Exception;
 
+        // Priority update
+        void priorityUpdate(int prioritizedStreamID, Priority p) throws Http2Exception;
+
         /**
          * Notification triggered when the parser swallows the payload of an unknown frame.
          *
@@ -729,5 +826,7 @@ class Http2Parser {
          * @throws IOException If an I/O occurred while swallowing the unknown frame
          */
         void onSwallowedUnknownFrame(int streamId, int frameTypeId, int flags, int size) throws IOException;
+
+        void increaseOverheadCount(FrameType frameType);
     }
 }
